




说明: 本笔记用于记录自己的买卖心得.
    1. 为什么买这只股票, 为什么卖这只股票?
        有这样的记录, 方便之后复盘, 否则自己都忘了操作的原因, 后期总结得失的时候, 总觉得像是靠猜靠运气, 并不是真正的实力.
        (比如之前操作的'科大讯飞', 后期看来是"神操作", 自己都佩服... 但是我也不太记得当初为啥有这个判断? 那么就没办法给未来的判断提供有价值的经验)
    2. 每一笔'买卖操作', 都需要强迫自己当天就记录在这个笔记中, 写下原因!! [强制性]
        通过'强制性'增加'麻烦', 来减少自己的操作....或许也是一件有利的事




[策略1]: 双均线(DoubleMovingAverage)
    使用场景:
        1. 股票类别: A股
        2. 经常会出现: 明显上涨或者下降通道的股票 (一般"当前业绩好"或者"未来前景好"的股票, 在利好消息下会出现这种情况)
                    (海康威视/科大讯飞/宁波银行/中国平安 从"长线角度"看, 其实都挺适合这种方式... )
        3. 适合做'长线操作' (即: 不需要太多买卖操作, 由市场蓬勃发展来赚钱) # 以月为单位?
    特点:
        1. 盈亏不太稳定, 但是涨幅还是不错的, 形势好的时候能赚的比'补损策略'多好多 [高风险高收益]


[策略2]: 补损(StopLoss)
    使用场景:
        1. 股票类别: ETF (没有交易手续费, 只有托管费, 所以比较适合高频交易, 赚快钱)
        2. 长期处于'横盘', '上下震荡', '没有明显上行下行趋势'的股票
                    ( 不太在乎它的基本面怎么样, 只求稳定的上下震荡. 做到高抛低吸的频繁交易)
                    ( 比如 惠城科技/大康农业 从"短线角度"看, 其实都挺适合这种方式... )
        3. 适合做'短线操作' (即: 需要频繁地主动买卖操作, 自身没有上涨空间, 但是可以赚取"波动的价差") # 以天为单位?
        4. 适合'涨跌幅度较大'的'创业板'/'科创板'.
            (幅度最好在5%-20%之间, 长期维持在2%左右也不行, 因为这个ETF搭配这个策略赚的是'波动的价差', 没有波动的ETF, 只是长期付托管费, 没投资意义)
    特点:
        1. 盈亏较为稳定, 但是涨幅不高, 形势好的时候能赚的比'补损策略'少, 但是基本上'不太会亏钱'... [保本小赚]


[时间线记录]
20200820:
    B:
        1. 科大讯飞: 新购100股
            长线视角:
                i. 第一季度净利润为负, 同比下降很多. 市盈率为负.... 市净率为7.5
                ii. 市值行业第6, 收入行业5, 净利润行业倒数7... (今年股价不涨确实是有原因的, 但未来看好)
                ii. 但是信仰加持 (非常看好它的人工智能的前沿科技)
                iii. 月K: 横盘, 今年没有明显上涨, 说明受疫情影响的业绩下滑体现在股市上就是'不涨'.
                          一旦疫情结束, 业绩利好, 相信会进入持续上涨通道 # (类似特斯拉/蔚来/亚马逊, 一旦扭亏为盈, 前途不可限量)
                    # 感觉可以把'月K的视角'划入'长线范畴'
            短线视角:
                i. 触及BOLL线的中轨. (连跌3天)
                ii. 科大讯飞近期处于'横盘震荡期', 可以赚这种'波动价差', 现在处于较低位, 可以低吸, 等反弹后高抛
            so: 近期可以做短线, 赚'波动价差', 之后可以转做长期... # 非常看好的股票

        2. 格力电器: 补仓200股
            长线视角:
                i.
            短线视角:
                i. 连阴了好几天, 出现"三只乌鸦".....
                    (正常来说是短线上'走弱'的强烈信号, 但是我不太相信这种说法. 我更愿意使用'StopLoss'策略来补仓)
                ii. 穿破了BOLL下轨 (不知道明天能不能反弹...但是从过去半个月的走势来看, 即使反弹也没啥力道. 所以更多的是为了长线持有...)
            so: 原持仓100股, 赶紧补了200股, 把整体盈亏拉低在1%左右.

        3. 军工行业ETF: 新购5100股
            长线视角:
                i. 没啥基本面需要看, 就看行业趋势和未来前景. (军工肯定不是夕阳产业, 问题不太大)
                ii. 行业性质: 它符合'ETF+补仓策略'的'涨跌幅度'条件
            短线视角:
                i. 日K:
                    1. 趋势线: 触及下趋势线
                    2. 布林线: 触及中轨均线
                        (都是反弹的信号)
                ii. 月K: 涨的有点高了...但是最近A股行情就是比较好, 处于高位也能接受, 如果A股突破3500大关, 后期还是比较看好.
            so: 先买了5100股试试水. (先上车再说, 说不定是通过天堂的车)
                # (另外的1手只花了100块,竟然在盘口资金上就跳出来了, 我在ETF中也是大户吗? 还是错觉??)

        4. 中国平安: 77元委托100股
            长线视角:
                i. 市盈率14, 市净率2. (类似银行股被低估) (大亏肯定不太会, 但也不会大涨)
                ii. 历史收入利润都不错, 今年收疫情影响利润同比下降40%.
                    (平安历年的利润率也都低于行业平均, 但比另一个龙头'中国人寿'强) (所以利润率低, 是规模做大后的正常现象??)
                ii. 保险股龙头 (历史走势都不错)
                iv. 净资产收益率20%左右, 总资产收益率2%左右...  (负债占比挺高的...)
            短线视角:
                i. 日K: 先是两天保险板块大涨, 然后连阴3天, 出现'三只乌鸦'...
                ii. 触及BOLL线的中轨.
            so: 有可能反弹, 但不是很确定, 所以以'低于市场价'委托, 期望等未来行情进一步明确后再下手 [观望]

    S:
        1. 长江电力: 100股全仓卖出
            长线视角:
                i. 虽然它的净利率是高于行业平均, 但是总资产收益率今年来看表现很差,甚至低于行业平均.
                ii. 利润同比下降20%(受疫情影响)
                iii. 市盈率竟然是行业垫底的部分(过高)...
            短线视角:
                i. 连涨了两天了, 触及了BOLL上轨
            other:
                i. 我也有点忘了, 昨天为啥买它.... 反正今天看来, 不管是'长线角度'还是'短线角度', 都不太适合再持有它...
                    (之前买入的原因是"MACD"上的量能线的'绿转红'...目前来看虽然'绿转红',但穿透了BOLL线的上轨..还是挺危险的...)
                ii. 今天昨天是电力板块大涨时期, 我在涨幅0.8%时候卖出, 没多久就涨到1.8%.... (有主力拉升?)



20200821:
    B:
        1. 大康农业: 新购1000股
            长线视角:
                i. 业绩相对同行偏低, 历年利润较差, 今年收入维持稳定, 市盈率175较高
                ii. 月K: 处于中低位, 15年7月最高6.6元 (当前2.9元)
            短线视角:
                i. 周K: 涨幅较大, 处于近期的高位
                ii. 日K: 大康很奇葩, 大盘跌的时候涨, 大盘涨的时候它跌
            so: 今天大盘大涨, 大康又跌了, 而且连跌2天, 先买1000股试试水. (也想着把历史的-20元给拉平了...)


    S:
        1. 军工行业ETF: 5100股全仓卖出
            长线视角:
                i. 见昨日买入时候的长线理由. (就是符合'ETF+补损策略')
            短线视角:
                i. BOLL: 昨日触及中轨, 预期会反弹; 今天冲高了一段时间, 而后跌破中轨, 预期持续下跌...
                ii. 趋势线: 昨日触及下趋势线, 预期会反弹; 今天冲高了一段时间, 而后跌破下趋势线, 预期持续下跌...
            so: 今日确有反弹, 但反弹1.5%, 幅度过高. 根基不稳, 担心没涨多久就下跌了, 赶紧出手, 锁定30元的盈利. [捂脸]

        2. 科大讯飞: 100股全仓卖出
            长线视角:
                i. 见0820买入时候的长线理由. (适合做长线战略部署)
            短线视角:
                i. 在连跌3天后, 感觉已经处于低位, 在昨日底部买入
                ii. 今日大涨4%, 触及BOLL线上轨, 预期下跌. (通过设定卖点交易, 而且刚好是最高的时候交易)
                    (我在"科大讯飞"上的操作简直就是神操作, 基本都在这'震荡市'中做到了'高抛低吸'....自己都佩服)

















































文档说明:
    1. 简写变量的解释说明:
        stock_code: "0.000002" (dc中叫'secid') # 获取行情数据时会使用
        sim_stock_code: "000002" (dc中叫'Zqdm') # 获取个人信息时会使用
        chicang_stock_code_lst: ["0.002505", "1.600685"]
        chicang_stock_count: 100  # 持仓数:100股
        cash_balance:
        init_cash_balance:
        deal_count:
        chicang_stock_today_market_vaule:
        chicang_stock_newest_market_vaule:
        chicang_stock_cost_market_vaule:
        total_capital_value:   # 总zz







    2. dc的简写变量名的对应关系:
        kqzj: 可取资金
        "Zqsl":证券数量, "Kysl":可用数量
        "Ckcb":xx成本, "Zxsz":最新市值
        ljyk:累计盈亏
        mmlb: 买卖类别









    end_price:
    JRZF: 今日
    growth_amount: 与上一个单位的差值
    growth_rate: 与上一个单位的增长率






















    todo:
        1. 通过现金流量表, 自己测算出整体的财务状况 [ok!!] [已完成]
            (东财在asset中显示的成本价很坑, 不是我想要的计算逻辑!!!)
                1. 持有1000股, 今天卖掉900股, 剩下的100股会改变'成本价'..... (神经吧)
                2. 同一天内, 早上把持有的100股卖了, 下午大跌所以又买回来200股.(正常来说, 早上的这100股的收益是确定的)
                    (但是东财把这3次买卖混在一起了. 混淆了我第二次下午购买的实际成本价了)
                    (成本价被摊低, 此次的浮盈比例也全乱了, 会让我错误估计当前股价是否该买...) [巨坑!!!]






##
